import numpy as np
import re
from typing import Iterable

from .lib_vars import *
from .lib_elements import *

#--------------------------------------------------------------------------

class Command:
    def __init__(self, name, inputs, outputs = None):
        self.name = name
        self.inputs = toObjArray(inputs)        # могут храниться сами объекты, а также строки с именами или выражениями
        self.outputs = toObjArray(outputs)

    def __repr__(self):
        inputs_str = ' '.join([obj.name if hasattr(obj, "name") else str(obj) for obj in self.inputs])
        outputs_str = ' '.join([obj.name if hasattr(obj, "name") else str(obj) for obj in self.outputs])
        return "{}:\t{}\t>> {}".format(self.name, inputs_str, outputs_str)

    def checkParamsPrepaired(self, params):
        for obj in params:
            if type(obj) not in type_to_shortcut:
                #raise Exception("In command <{}> param ({} as {}) is not prepaired.".format(self.name, obj, type(obj)))
                return False
        return True

    def func(self):     # эта функция должна находить реализованную функцию в этом модуле и возвращать ссылку на нее
        #if not self.checkParamsPrepaired(self.inputs):
        #    raise Exception("In command <{}> inputs are not prepaired.".format(self.name))

        fname = strFullCommand(self.name, self.inputs)

        if fname in globals(): return globals()[fname]
        else: return None

#--------------------------------------------------------------------------        

def toObjArray(params):
    if not params: return []
    elif not isinstance(params, Iterable) or isinstance(params, str): return [params]
    else: return list(params)

#--------------------------------------------------------------------------   

type_to_shortcut = {
    Point           : 'p',
    Polygon         : 'P',
    Circle          : 'c',
    Arc             : 'C',
    Line            : 'l',
    Ray             : 'r',
    Segment         : 's',
    Angle           : 'a',
    Vector          : 'v',
    CircleSector    : 'S',

    int             : 'i',
    float           : 'i',
    Boolean         : 'b',
    Measure         : 'm',
    AngleSize       : 'A'
}

def strFullCommand(name, params): # "Polygon", [Point, Point, int] -> "polygon_ppi"
    # print(f"strFullCommand: {name} - {params}")
    if strCommand(name) == "polygon" and bool(re.match('^p*$', strParams(params))):
        return strCommand(name)
    else:
        return f"{strCommand(name)}_{strParams(params)}"

def strParams(params): # [Point, Point, int] -> "ppi"
    for x in params:
        if type(x) not in type_to_shortcut:
            if x is not None:
                print(f"WARNING: no implementation for type '{type(x)}'\n")
    return ''.join(type_to_shortcut[type(x)] if type(x) in type_to_shortcut else '?' for x in params)

def strCommand(name): # "AreCollinear" -> "are_collinear"
    altered_name = [name[0].lower()]
    for x in name[1:]:
        if x.isupper(): altered_name += ['_', x.lower()]
        else: altered_name.append(x)
    return ''.join(altered_name)

#--------------------------------------------------------------------------

def res_by_index(res, index, points_order=None):
    index = int(max(0, index-1))
    
    if not hasattr(res, '__len__'):
        if index == 0: return res
        else: return None
        
    if len(res) > index: return res[len(res) - index - 1]
    else: return None
    
def res_by_index(res, index, points_order=None):
    index = int(max(0, index - 1))
    
    if not hasattr(res, '__len__'):
        return res if index == 0 else None

    if points_order is None:
        return res[index] if index < len(res) else None

    # Создаем упорядоченный список точек
    ordered_points = []
    used = [False] * len(res)  # Отслеживаем использованные точки из res
    
    # Сначала добавляем точки из points_order в указанном порядке
    for order_point in points_order:
        for i, res_point in enumerate(res):
            if not used[i] and np.isclose(order_point.a, res_point.a).all():
                ordered_points.append(res_point)
                used[i] = True
                break

    # Затем добавляем оставшиеся точки из res в исходном порядке
    for i, res_point in enumerate(res):
        if not used[i]:
            ordered_points.append(res_point)

    return ordered_points[index] if index < len(ordered_points) else None
    
#--------------------------------------------------------------------------

def assign_p(x):
    return x
def assign_v(x):
    return x
def assign_i(x):
    return x

def angle_size_A(a):
    return AngleSize(a.value)

def angle_size_i(x):
    return AngleSize(x)

def angle_ppp(p1, p2, p3):
    return Angle(p2.a, p1.a-p2.a, p3.a-p2.a)

def angle_size_ppp(p1, p2, p3):
    return AngleSize(np.angle(a_to_cpx(p3.a-p2.a) / a_to_cpx(p1.a-p2.a)))

def angular_bisector_ll(l1, l2):
    x = intersect_ll(l1, l2)
    n1, n2 = l1.n, l2.n
    if np.dot(n1, n2) > 0: n = n1 + n2
    else: n = n1 - n2
    return [
        Line(vec, np.dot(vec, x.a))
        for vec in (n, vector_perp_rot(n))
    ]

def angular_bisector_ppp(p1, p2, p3):
    v1 = p2.a - p1.a
    v2 = p2.a - p3.a
    v1 /= np.linalg.norm(v1)
    v2 /= np.linalg.norm(v2)
    if np.dot(v1, v2) < 0: n = v1-v2
    else: n = vector_perp_rot(v1+v2)
    return Line(n, np.dot(p2.a, n))

def angular_bisector_ss(l1, l2):
    return angular_bisector_ll(l1, l2)

def arc_cpp(circle, p1, p2):
    if np.isclose(circle.c, p1.a).all(): return None
    if np.isclose(circle.c, p2.a).all(): return None
    return Arc(circle.c, circle.r, [np.arctan2(p.a[1] - circle.c[1], p.a[0] - circle.c[0]) for p in (p1, p2)])

def circumcircle_arc_ppp(p1, p2, p3):
    if np.isclose(p1.a, p2.a).all(): return None
    if np.isclose(p1.a, p3.a).all(): return None
    if np.isclose(p2.a, p3.a).all(): return None
    circle = circle_ppp(p1, p2, p3)
    arc = Arc(circle.c, circle.r, [np.arctan2(p.a[1] - circle.c[1], p.a[0] - circle.c[0]) for p in (p1, p3)])
    return arc if arc.contains(p2.a) else Arc(circle.c, circle.r, [np.arctan2(p.a[1] - circle.c[1], p.a[0] - circle.c[0]) for p in (p3, p1)])

def circle_arc_ppp(p1, p2, p3):
    if np.isclose(p1.a, p2.a).all(): return None
    if np.isclose(p1.a, p3.a).all(): return None
    if np.isclose(p2.a, p3.a).all(): return None
    circle = circle_pp(p1, p2)
    return Arc(circle.c, circle.r, [np.arctan2(p.a[1] - circle.c[1], p.a[0] - circle.c[0]) for p in (p2, p3)])

def circumcircle_sector_ppp(p1, p2, p3):
    if np.isclose(p1.a, p2.a).all(): return None
    if np.isclose(p1.a, p3.a).all(): return None
    if np.isclose(p2.a, p3.a).all(): return None
    circle = circle_ppp(p1, p2, p3)
    arc = Arc(circle.c, circle.r, [np.arctan2(p.a[1] - circle.c[1], p.a[0] - circle.c[0]) for p in (p1, p3)])
    return CircleSector(arc.c, arc.r, arc.angles) if arc.contains(p2.a) else CircleSector(arc.c, arc.r, list(reversed(arc.angles)))

def circle_sector_ppp(p1, p2, p3):
    if np.isclose(p1.a, p2.a).all(): return None
    if np.isclose(p1.a, p3.a).all(): return None
    if np.isclose(p2.a, p3.a).all(): return None
    circle = circle_pp(p1, p2)
    return CircleSector(circle.c, circle.r, [np.arctan2(p.a[1] - circle.c[1], p.a[0] - circle.c[0]) for p in (p2, p3)])

def circle_sector_ppA(p1, p2, ang):
    if np.isclose(p1.a, p2.a).all(): return None
    circle = circle_pp(p1, p2)
    ang0 = np.arctan2(p2.a[1] - circle.c[1], p2.a[0] - circle.c[0])
    return CircleSector(circle.c, circle.r, [ang0, ang0 + ang.value])

def circle_sector_ppi(p1, p2, value):
    if np.isclose(p1.a, p2.a).all(): return None
    circle = circle_pp(p1, p2)
    ang0 = np.arctan2(p2.a[1] - circle.c[1], p2.a[0] - circle.c[0])
    return CircleSector(circle.c, circle.r, [ang0, ang0 + value])

def are_collinear_ppp(p1, p2, p3):
    return Boolean(np.linalg.matrix_rank([p1.a-p2.a, p1.a-p3.a]) <= 1)

def are_concurrent_lll(l1, l2, l3):
    lines = l1,l2,l3
    
    differences = []
    for i in range(3):
        remaining = [l.n for l in lines[:i]+lines[i+1:]]
        prod = np.abs(np.cross(*remaining))
        differences.append((prod, i, lines[i]))

    l1, l2, l3 = tuple(zip(*sorted(differences)))[2]
    x = intersect_ll(l1, l2)
    return Boolean(np.isclose(np.dot(x.a, l3.n), l3.c))

def are_concurrent(o1, o2, o3):
    cand = []
    try:
        #if True:
        if isinstance(o1, Line) and isinstance(o2, Line):
            cand = intersect_ll(o1, o2)
        elif isinstance(o1, Line) and isinstance(o2, Circle):
            cand = intersect_lc(o1, o2)
        elif isinstance(o1, Circle) and isinstance(o2, Line):
            cand = intersect_cl(o1, o2)
        elif isinstance(o1, Circle) and isinstance(o2, Circle):
            cand = intersect_cc(o1, o2)
    except: pass

    if not isinstance(cand, Iterable) or isinstance(cand, str): cand = (cand,)

    for p in cand:
        for obj in (o1,o2,o3):
            if not obj.contains(p.a): break
        else: return Boolean(True)

    return Boolean(False)

def are_concyclic_pppp(p1, p2, p3, p4):
    z1, z2, z3, z4 = (a_to_cpx(p.a) for p in (p1, p2, p3, p4))
    cross_ratio = (z1-z3)*(z2-z4)*(((z1-z4)*(z2-z3)).conjugate())
    return Boolean(np.isclose(cross_ratio.imag, 0))

def are_congruent_aa(a1, a2):
    result = np.isclose((a1.angle-a2.angle+1)%(2*np.pi), 1)
    result = (result or np.isclose((a1.angle+a2.angle+1)%(2*np.pi), 1))
    return Boolean(result)

def are_complementary_aa(a1, a2):
    result = np.isclose((a1.angle-a2.angle)%(2*np.pi), np.pi)
    result = (result or np.isclose((a1.angle+a2.angle)%(2*np.pi), np.pi))
    return Boolean(result)

def are_congruent_ss(s1, s2):
    l1, l2 = (
        np.linalg.norm(s.end_points[1] - s.end_points[0])
        for s in (s1, s2)
    )
    return Boolean(np.isclose(l1, l2))

def are_equal_mm(m1, m2):
    if m1.dim != m2.dim: return None
    return Boolean(np.isclose(m1.x, m2.x))

def are_equal_mi(m, i):
    if m.dim != 0: return None
    return Boolean(np.isclose(m.x, i))

def are_equal_pp(p1, p2):
    return Boolean(np.isclose(p1.a, p2.a).all())

def are_parallel_ll(l1, l2):
    if np.isclose(l1.n, l2.n).all(): return Boolean(True)
    if np.isclose(l1.n, -l2.n).all(): return Boolean(True)
    return Boolean(False)

def are_parallel_ls(l, s):
    return are_parallel_ll(l, s)

def are_parallel_rr(r1, r2):
    return are_parallel_ll(r1, r2)

def are_parallel_sl(s, l):
    return are_parallel_ll(s, l)

def are_parallel_ss(s1, s2):
    return are_parallel_ll(s1, s2)

def are_perpendicular_ll(l1, l2):
    if np.isclose(l1.n, l2.v).all(): return Boolean(True)
    if np.isclose(l1.n, -l2.v).all(): return Boolean(True)
    return Boolean(False)

def are_perpendicular_lr(l, r):
    return are_perpendicular_ll(l, r)

def are_perpendicular_rl(r, l):
    return are_perpendicular_ll(r, l)

def are_perpendicular_sl(s, l):
    return are_perpendicular_ll(s, l)

def are_perpendicular_ls(l, s):
    return are_perpendicular_ll(l, s)

def are_perpendicular_ss(s1, s2):
    return are_perpendicular_ll(s1, s2)

def area(*points):
    p0 = points[0].a
    vecs = [p.a - p0 for p in points[1:]]
    cross_sum = sum(
        np.cross(v1, v2)
        for v1, v2 in zip(vecs, vecs[1:])
    )
    return Measure(abs(cross_sum)/2, 2)

def area_P(polygon):
    points = [Point(p) for p in polygon.points]
    return area(*points)

def center_c(circle):
    return Point(circle.c)

def centroid_P(polygon):
    """Вычисляет центроид многоугольника и возвращает его как объект Point."""
    points = polygon.points
    n = len(points)
    
    # Если многоугольник имеет менее 3 точек, возвращаем среднее арифметическое
    if n < 3:
        centroid_coord = np.mean(points, axis=0)
        return Point(centroid_coord)
    
    # Вычисляем площадь по формуле шнурка
    area = 0
    for i in range(n):
        x1, y1 = points[i]
        x2, y2 = points[(i + 1) % n]
        area += x1 * y2 - x2 * y1
    area /= 2
    
    # Если площадь близка к нулю, возвращаем среднее арифметическое точек
    if abs(area) < 1e-10:
        centroid_coord = np.mean(points, axis=0)
        return Point(centroid_coord)
    
    # Вычисляем центроид
    Cx = Cy = 0
    for i in range(n):
        x1, y1 = points[i]
        x2, y2 = points[(i + 1) % n]
        cross = x1 * y2 - x2 * y1
        Cx += (x1 + x2) * cross
        Cy += (y1 + y2) * cross
    
    Cx /= 6 * area
    Cy /= 6 * area
    
    return Point([Cx, Cy])

def circle_pp(center, passing_point):
    return Circle(center.a, np.linalg.norm(center.a - passing_point.a))

def circle_ppp(p1, p2, p3):
    axis1 = line_bisector_pp(p1, p2)
    axis2 = line_bisector_pp(p1, p3)
    center = intersect_ll(axis1, axis2)
    return circle_pp(center, p1)

def circle_pm(p, m):
    if m.dim != 1: return None
    return Circle(p.a, m.x)

def circle_pi(p, i):
    return Circle(p.a, i)

def circle_ps(p, s):
    return Circle(p.a, s.length)

def contained_by_pc(point, by_circle):
    return Boolean(by_circle.contains(point.a))

def contained_by_pl(point, by_line):
    return Boolean(by_line.contains(point.a))

def distance_pp(p1, p2):
    return Measure(np.linalg.norm(p1.a-p2.a), 1)

def equality_aa(a1, a2):
    return are_congruent_aa(a1, a2)

def equality_mm(m1, m2):
    if m1.dim != m2.dim: return None
    return Boolean(np.isclose(m1.x, m2.x))

def equality_ms(m, s):
    if m.dim != 1: return None
    return Boolean(np.isclose(m.x, s.length))

def equality_mi(m, i):
    if m.dim != 0 and i != 0: return None
    return Boolean(np.isclose(m.x, i))

def equality_pp(p1, p2):
    return are_equal_pp(p1, p2)

def equality_Pm(polygon, m):
    if m.dim != 2: return None
    return Boolean(np.isclose(area_P(polygon).x, m.x))

def equality_PP(poly1, poly2):
    return Boolean(np.isclose(area_P(poly1).x, area_P(poly2).x))

def equality_sm(s, m):
    return equality_ms(m,s)

def equality_ss(s1, s2):
    return Boolean(np.isclose(s1.length, s2.length))

def equality_si(seg, a):
    return Boolean(np.isclose(seg.length, a))

def intersect_ll(line1, line2):
    matrix = np.stack((line1.n, line2.n))
    b = np.array((line1.c, line2.c))
    if np.isclose(np.linalg.det(matrix), 0): return None
    return Point(np.linalg.solve(matrix, b))

def intersect_lc(line, circle):
    # shift circle to center
    y = line.c - np.dot(line.n, circle.c)
    x_squared = circle.r_squared - y ** 2
    if np.isclose(x_squared, 0): return Point(y * line.n + circle.c)
    if x_squared <= 0: return None

    x = np.sqrt(x_squared)
    return [
        Point(-x * line.v + y * line.n + circle.c),
        Point(x * line.v + y * line.n + circle.c)
    ]

def intersect_lci(line, circle, index, points_order=None):
    res = intersect_lc(line, circle)
    return res_by_index(res, index, points_order)

def intersect_cc(circle1, circle2):
    center_diff = circle2.c - circle1.c
    center_dist_squared = np.dot(center_diff, center_diff)
    # center_dist = np.sqrt(center_dist_squared)
    relative_center = (circle1.r_squared - circle2.r_squared) / center_dist_squared
    center = (circle1.c + circle2.c) / 2 + relative_center*center_diff / 2

    rad_sum  = circle1.r + circle2.r
    rad_diff = circle1.r - circle2.r
    det = (rad_sum ** 2 - center_dist_squared) * (center_dist_squared - rad_diff ** 2)
    if np.isclose(det, 0): return Point(center)
    if det <= 0: return None
    center_deviation = np.sqrt(det)
    center_deviation = np.array(((-center_deviation,),(center_deviation,)))

    return [
        Point(center + center_dev)
        for center_dev in center_deviation * 0.5 * vector_perp_rot(center_diff) / center_dist_squared
    ]

def intersect_cci(circle1, circle2, index, points_order=None):
    res = intersect_cc(circle1, circle2)
    return res_by_index(res, index, points_order)
    
def intersect_cl(circle, line):
    return intersect_lc(line, circle)

def intersect_cli(circle, line, index, points_order=None):
    return intersect_lci(line, circle, index, points_order)

def intersect_cr(circle, ray):
    return intersect_rc(ray, circle)

def intersect_cri(circle, ray, index, points_order=None):
    return intersect_rci(ray, circle, index, points_order)

def intersect_Cl(arc, line):
    results = intersect_lc(line,arc)
    if not isinstance(results, Iterable) or isinstance(results, str): results = (results,)
    return [p for p in results if arc.contains(p.a)]

def intersect_lC(line, arc):
    return intersect_Cl(arc, line)

def intersect_Cli(arc, line, index, points_order=None):
    res = intersect_Cl(arc, line)
    return res_by_index(res, index, points_order)

def intersect_lCi(line, arc, index, points_order=None):
    return intersect_Cli(arc, line, index, points_order)

def intersect_Sl(sector, line):
    return intersect_Cl(Arc(sector.c, sector.r, sector.angles), line)

def intersect_lS(line, sector):
    return intersect_Sl(sector, line)

def intersect_Sli(sector, line, index, points_order=None):
    res = intersect_Sl(sector, line)
    return res_by_index(res, index, points_order)

def intersect_lSi(line, sector, index, points_order=None):
    return intersect_Sli(sector, line, index, points_order)

def intersect_cs(circle, segment):
    results = intersect_lc(segment, circle)
    if (not isinstance(results, Iterable)) or isinstance(results, str): results = (results,)
    return [p for p in results if segment.contains(p.a)]

def intersect_csi(circle, segment, index, points_order=None):
    res = intersect_cs(circle, segment)
    return res_by_index(res, index, points_order)

def intersect_rc(ray, circle):
    results = intersect_lc(ray, circle)
    if (not isinstance(results, Iterable)) or isinstance(results, str): results = (results,)
    return [p for p in results if ray.contains(p.a)]

def intersect_rci(ray, circle, index, points_order=None):
    res = intersect_rc(ray, circle)
    return res_by_index(res, index, points_order)

def intersect_sc(segment, circle):
    return intersect_cs(circle, segment)

def intersect_sci(segment, circle, index, points_order=None):
    return intersect_csi(circle, segment, index, points_order)

def intersect_lr(line, ray):
    result = intersect_ll(line, ray)
    if result is None: return None
    return result if ray.contains(result.a) else None

def intersect_ls(line, segment):
    result = intersect_ll(line, segment)
    if result is None: return None
    return result if segment.contains(result.a) else None

def intersect_rl(ray, line):
    return intersect_lr(line, ray)

def intersect_rr(r1, r2):
    result = intersect_ll(r1, r2)
    if result is None: return None
    if not r1.contains(result.a): return None
    if not r2.contains(result.a): return None
    return result

def intersect_rs(ray, segment):
    result = intersect_ll(ray, segment)
    if result is None: return None
    if not ray.contains(result.a): return None
    if not segment.contains(result.a): return None
    return result

def intersect_sl(segment, line):
    return intersect_ls(line, segment)

def intersect_sr(segment, ray):
    return intersect_rs(ray, segment)

def intersect_ss(s1, s2):
    result = intersect_ll(s1, s2)
    if result is None: return None
    if not s1.contains(result.a): return None
    if not s2.contains(result.a): return None
    return result

def line_bisector_pp(p1, p2):
    p = (p1.a + p2.a) / 2
    n = p2.a - p1.a
    if (n == 0).all(): return None
    return Line(n, np.dot(n, p))

def line_bisector_s(segment):
    p1, p2 = segment.end_points
    p = (p1 + p2) / 2
    n = p2 - p1
    return Line(n, np.dot(n, p))

def line_pl(point, line):
    return Line(line.n, np.dot(line.n, point.a))

def line_pp(p1, p2):
    if (p1.a == p2.a).all(): return None
    n = vector_perp_rot(p1.a - p2.a)
    return Line(n, np.dot(p1.a, n))

def line_pv(pt, vec):
    if (vec.v == 0).all(): return None
    n = vector_perp_rot(vec.v)
    return Line(n, np.dot(pt.a, n))

def line_pr(point, ray):
    return line_pl(point, ray)

def line_ps(point, segment):
    return line_pl(point, segment)

def line_s(segment):
    return Line(segment.n, segment.c)

def midpoint_pp(p1, p2):
    return Point((p1.a+p2.a)/2)

def midpoint_s(segment):
    p1, p2 = segment.end_points
    return Point((p1 + p2) / 2)

def mirror_cc(circle, by_circle):
    center_v = circle.c - by_circle.c
    denom = square_norm(center_v) - circle.r_squared
    if np.isclose(denom, 0):
        return Line(center_v, circle.r_squared / 2 + np.dot(center_v, by_circle.c))
    else:
        return Circle(
            center = (by_circle.r_squared/denom) * center_v + by_circle.c,
            r = by_circle.r_squared * circle.r / abs(denom)
        )

def mirror_cl(circle, by_line):
    return Circle(
        center = circle.c + by_line.n*2*(by_line.c - np.dot(circle.c, by_line.n)),
        r = circle.r,
    )

def mirror_cp(circle, by_point):
    return Circle(
        center = 2 * by_point.a - circle.c,
        r = circle.r
    )

def mirror_ll(line, by_line):
    n = line.n - by_line.n * 2 * np.dot(line.n, by_line.n)
    return Line(n, line.c + 2 * by_line.c * np.dot(n, by_line.n) )

def mirror_ls(line, by_segment):
    return mirror_ll(line, by_segment)

def mirror_lp(line, by_point):
    return Line(line.n, 2 * np.dot(by_point.a, line.n) - line.c)

def mirror_pc(point, by_circle):
    v = point.a - by_circle.c
    if np.isclose(v, 0).all(): return None
    return Point(by_circle.c + v * (by_circle.r_squared / square_norm(v)) )

def mirror_pl(point, by_line):
    return Point(point.a + by_line.n * 2 * (by_line.c - np.dot(point.a, by_line.n)))

def mirror_pp(point, by_point):
    return Point(2 * by_point.a - point.a)

def mirror_ps(point, by_segment):
    return mirror_pl(point, by_segment)

def orthogonal_line_pl(point, line):
    return Line(-line.v, np.dot(-line.v, point.a))

def orthogonal_line_pr(point, ray):
    return orthogonal_line_pl(point, ray)

def orthogonal_line_ps(point, segment):
    return orthogonal_line_pl(point, segment)

def point_():
    return Point(np.random.normal(size = 2))

def point_ii(x, y):
    return Point([x,y])

def point_c(circle, alpha = None):
    return Point(circle.c + circle.r * get_direction(alpha))

def point_l(line, alpha = None):
    return Point(line.c * line.n + line.v * (alpha if alpha is not None else np.random.normal()) )

def point_s(segment, alpha = None):
    return Point(interpolate(segment.end_points[0], segment.end_points[1], alpha if alpha is not None else np.random.random()))

def polar_pc(point, circle):
    n = point.a - circle.c
    if np.isclose(n, 0).all(): return None
    return Line(n, np.dot(n, circle.c) + circle.r_squared)

def polygon_ppi(p1, p2, n):
    p1c, p2c = (a_to_cpx(p.a) for p in (p1, p2))
    alpha = 2 * np.pi / n
    center = p2c + (p1c - p2c) / (1 - np.exp(-alpha * 1j))
    v = p2c - center
    points = [Point(cpx_to_a(center + v * np.exp(i * alpha * 1j))) for i in range(1, int(n) - 1)]
    raw_points = [p.a for p in [p1, p2] + points]
    segments = [
        Segment(p1, p2)
        for p1, p2 in zip(raw_points, raw_points[1:] + raw_points[:1])
    ]
    return [Polygon(raw_points)] + segments + points

def polygon(*points):
    raw_points = [p.a for p in points]
    segments = [
        Segment(p1, p2)
        for p1, p2 in zip(raw_points, raw_points[1:] + raw_points[:1])
    ]
    return [Polygon(raw_points)] + segments

def prove_b(x):
    print(x.b)
    return x

def radius_c(circle):
    return Measure(circle.r, 1)

def ray_pp(p1, p2):
    return Ray(p1.a, p2.a - p1.a)

def rotate_pap(point, angle, by_point):
    return Point(by_point.a + rotate_vec(point.a - by_point.a, angle.angle))

def rotate_pAp(point, angle_size, by_point):
    return Point(by_point.a + rotate_vec(point.a - by_point.a, angle_size.value))

def rotate_pip(point, angle_value, by_point):
    return Point(by_point.a + rotate_vec(point.a - by_point.a, angle_value))

def rotate_vAp(vec, angle_size, by_point):
    return Vector([vec.end_points[0], vec.end_points[0] + rotate_vec(vec.v, angle_size.value)])

def segment_pp(p1, p2):
    return Segment(p1.a, p2.a)

def segment_pv(p, vec):
    return Segment(p.a, p.a + vec.v)

def semicircle_pp(p1, p2):
    vec = a_to_cpx(p1.a - p2.a)
    return Arc(
        (p1.a + p2.a) / 2,
        abs(vec) / 2,
        [np.angle(v) for v in (-vec, vec)]
    )

def tangent_pc(point, circle):
    polar = polar_pc(point, circle)
    intersections = intersect_lc(polar, circle)
    if isinstance(intersections, Iterable) and not isinstance(intersections, str) and len(intersections) == 2:
        return [line_pp(point, x) for x in intersections]
    else: return polar

def tangent_pci(point, circle, index):
    res = tangent_pc(point, circle)
    index = int(max(0, index-1))
    
    if not hasattr(res, '__len__'):
        if index == 0: return res
        else: return None
        
    if len(res) > index: return res[len(res) - index - 1]
    else: return None

def touches_cc(c1, c2):
    lens = c1.r, c2.r, np.linalg.norm(c1.c-c2.c)
    return Boolean(np.isclose(sum(lens), 2 * max(lens)))

def touches_lc(line, circle):
    return Boolean(
        np.isclose(circle.r, np.abs(np.dot(line.n, circle.c) - line.c) )
    )

def touches_cl(circle, line):
    return touches_lc(line, circle)

def translate_pv(point, vector):
    return Point(point.a + vector.v)

def translate_sv(segment, vector):
    return Segment(*(segment.end_points + vector.v))

def translate_cv(circle, vector):
    return Circle(circle.c + vector.v, circle.r)

def vector_pp(p1, p2):
    return Vector((p1.a, p2.a))

def vector_p(p1):
    return Vector(([0, 0], p1.a))

def vector_ii(x, y):
    return Vector(([0, 0], [x, y]))

def vector_vi(vec, mod):
    k = mod / np.linalg.norm(vec.v)
    return Vector(vec.end_points * k)

#--------------------------------------------------------

def abs_i(a):
    return np.abs(a)

def abs_m(m):
    return Measure(np.abs(m.x), m.dim)

def value_A(ang_size):
    return float(ang_size.value)

def u_sub_i(a):
    return float(-a)

def u_sub_v(vector):
    return Vector(-(vector.end_points))

def u_sub_p(point):
    return Point(-(point.a))

def u_sub_a(angle):
    return AngleSize(-angle.angle)

def u_sub_A(anglesize):
    return AngleSize(-anglesize.value)

def u_sub_m(m):
    return Measure(-m.x, m.dim)

def sub_ii(a, b):
    return float(a - b)

def sub_AA(a1, a2):
    return AngleSize(a1.value - a2.value)

def sub_aA(a, A):
    return AngleSize(a.angle - A.value)

def sub_Aa(A, a):
    return AngleSize(A.value - a.angle)

def sub_vv(v1, v2):
    return Vector(v1.end_points - v2.end_points)

def sub_pv(point, vector):
    return Point(point.a - vector.v)

def sub_pp(p1, p2):
    return Point(p1.a - p2.a)

def sub_mm(m1, m2):
    if m1.dim != m2.dim: return None
    return Measure(m1.x - m2.x, m1.dim)

def sub_ms(m, s):
    if m.dim != 1: return None
    return Measure(m.x - s.length, 1)

def sub_sm(s, m):
    if m.dim != 1: return None
    return Measure(s.length - m.x, 1)

def sub_ss(s1, s2):
    return Measure(s1.length - s2.length, 1)

def pow_ii(a, b):
    return float(a ** b)

def pow_mi(m, i):
    if i != 2: return None
    return Measure(m.x ** i, m.dim*i)

def pow_si(s, i):
    return Measure(s.length ** i, i)

def mult_mi(m, b):
    return float(m.x * b)

def mult_im(a, m):
    return float(a * m.x)

def mult_ii(a, b):
    return float(a * b)

def mult_vi(vector, a):
    return Vector(vector.end_points * a)

def mult_iv(a, vector):
    return mult_vi(vector, a)

def mult_ip(a, point):
    return Point(point.a * a)

def mult_pi(point, a):
    return Point(point.a * a)

def mult_mm(m1, m2):
    return Measure(m1.x * m2.x, m1.dim + m2.dim)

def mult_ms(m, s):
    return Measure(m.x * s.length, m.dim + 1)

def mult_mf(m, f):
    return Measure(m.x * f, m.dim)

def mult_sm(s, m):
    return mult_ms(m,s)

def mult_ss(s1, s2):
    return Measure(s1.length * s2.length, 2)

def mult_fm(f, m):
    return mult_mf(m, f)

def mult_ff(f1, f2):
    return Measure(f1 * f2, 0)

def mult_iA(i, angle_size):
    return AngleSize(angle_size.value * i)

def mult_Ai(a, i):
    return mult_iA(i, a)

def mult_is(i, s):
    return Measure(i * s.length, 1)

def mult_if(i, f):
    return Measure(i * f, 0)

def div_ii(a, b):
    if b == 0: return None
    return float(a / b)

def div_Ai(angle_size, i):
    return AngleSize(angle_size.value / i)

def div_mm(m1, m2):
    if np.isclose(m2.x, 0): return None
    return Measure(m1.x / m2.x, m1.dim - m2.dim)

def div_ms(m, s):
    return Measure(m.x / s.length, m.dim - 1)

def div_mi(m, i):
    if np.isclose(i, 0): return None
    return Measure(m.x / i, m.dim)

def div_sm(s, m):
    if np.isclose(m.x, 0): return None
    return Measure(s.length / m.x, 1 - m.dim)

def div_ss(s1, s2):
    return Measure(s1.length / s2.length, 0)

def div_si(s, i):
    if np.isclose(i, 0): return None
    return Measure(s.length / i, 1)

def div_vi(vector, a):
    return Vector(vector.end_points / a)

def div_pi(point, a):
    return Point(point.a / a)

def add_ii(a, b):
    return float(a + b)

def add_vv(v1, v2):
    return Vector(v1.end_points + v2.end_points)

def add_pv(point, vector):
    return Point(point.a + vector.v)

def add_vp(vector, point):
    return add_pv(point, vector)

def add_pp(p1, p2):
    return Point(p1.a + p2.a)

def add_mm(a, b):
    return float(a + b)

def add_mm(m1, m2):
    if m1.dim != m2.dim: return None
    return Measure(m1.x + m2.x, m1.dim)

def add_ms(m, s):
    if m.dim != 1: return None
    return Measure(m.x + s.length, 1)

def add_mi(m, i):
    if m.dim != 0: return None
    return Measure(m.x + i, 0)

def add_ss(s1, s2):
    return Measure(s1.length + s2.length, 1)

#--------------------------------------------------------

def cos_i(x): return float(np.cos(x))
def sin_i(x): return float(np.sin(x))
def tan_i(x): return float(np.tan(x))
def ctan_i(x): return float(1 / np.tan(x))
def sqrt_i(x): return float(np.sqrt(x))



import os
import shutil
import re
import ast
from zipfile import ZipFile
from xml.etree import ElementTree
from xml.etree.ElementTree import Element as XElement

from ..geo.construction import Construction
from ..geo.lib_commands import Command
from ..geo.lib_vars import *
from ..geo.lib_elements import *
from .short_parser import putCode  # Импорт функции для выполнения кода

temp_path = os.path.join(os.getcwd(), "temp")

#--------------------------------------------------------------------------

from lark import Lark, Transformer, v_args
import xml.etree.ElementTree as ET

# Включаем преобразование аргументов в узлы дерева
@v_args(inline=True)
class GeoGebraTransformer(Transformer):
    def __init__(self):
        super().__init__()
    
    def number(self, token):
        return float(token)
    
    def variable(self, token):
        return str(token)
    
    def coordinates(self, x, y):
        return ("coordinates", x, y)
    
    def function_call(self, name, *args):
        # Фильтруем запятые и None
        args = [arg for arg in args if arg not in (',', None)]
        return ("function_call", str(name), args)
    
    def add(self, left, right):
        return ("bin_op", "+", left, right)
    
    def sub(self, left, right):
        return ("bin_op", "-", left, right)
    
    def mul(self, left, right):
        return ("bin_op", "*", left, right)
    
    def div(self, left, right):
        return ("bin_op", "/", left, right)
    
    def pow(self, left, right):
        return ("bin_op", "^", left, right)
    
    def neg(self, expr):
        return ("unary_op", "-", expr)
    
    def expression(self, expr):
        return expr
    
    def start(self, expr):
        return expr

# Грамматика с правильным приоритетом операций
grammar = """
    start: expression

    expression: sum

    sum: product
        | sum "+" product -> add
        | sum "-" product -> sub

    product: power
           | product "*" power -> mul
           | product "/" power -> div

    power: atom
         | power "^" atom -> pow

    atom: number
        | variable
        | coordinates
        | function_call
        | "-" atom -> neg
        | "(" expression ")"

    number: SIGNED_NUMBER
    variable: CNAME

    coordinates: "(" expression "," expression ")"

    function_call: CNAME "[" [args] "]"  -> function_call
                 | CNAME "(" [args] ")"  -> function_call

    args: expression ("," expression)*

    %import common.CNAME
    %import common.SIGNED_NUMBER
    %import common.WS
    %ignore WS
"""

# Создаем парсер
lark_parser = Lark(grammar, start="start", parser="lalr", transformer=GeoGebraTransformer())

# Функция для обработки специальных символов (например, градусов)
def preprocess_expression(expr):
    # Заменяем специальные символы
    expr = expr.replace("°", "")  # Удаляем символ градуса
    # Можно добавить другие замены по необходимости
    return expr

# Функция для парсинга XML GeoGebra
def parse_geogebra_xml(xml_content):
    root = ET.fromstring(xml_content)
    results = []
    
    # Парсим выражения из команд
    for cmd in root.findall(".//command"):
        for inp in cmd.findall("input"):
            for attr_name, expr in inp.items():
                if attr_name.startswith('a'):
                    try:
                        processed_expr = preprocess_expression(expr)
                        parsed = lark_parser.parse(processed_expr)
                        results.append(("command_input", cmd.get("name"), attr_name, expr, parsed))
                    except Exception as e:
                        results.append(("error", f"Ошибка разбора {expr}: {e}"))
    
    # Парсим выражения из элементов expression
    for expr_elem in root.findall(".//expression"):
        expr = expr_elem.get("exp")
        if expr:
            try:
                processed_expr = preprocess_expression(expr)
                parsed = lark_parser.parse(processed_expr)
                results.append(("expression", expr_elem.get("label"), expr, parsed))
            except Exception as e:
                results.append(("error", f"Ошибка разбора {expr}: {e}"))
    
    return results

#--------------------------------------------------------------------------

def rgb_to_hex(r, g, b):
    return '#{:02x}{:02x}{:02x}'.format(r, g, b)

def get_xelems(ggb_path: str):
    try:    
        os.mkdir(temp_path)
    except FileExistsError:
        pass
    shutil.copyfile(ggb_path, os.path.join(temp_path, "temp.ggb"))
    
    ggb = ZipFile(os.path.join(temp_path, "temp.ggb"))
    ggb.extractall(temp_path)
    ggb.close()
    os.remove(os.path.join(temp_path, "temp.ggb"))
    
    tree = ElementTree.parse(os.path.join(temp_path, "geogebra.xml"))
    root = tree.getroot()
    
    shutil.rmtree(temp_path)
    
    return root.find("construction"), root.find("euclidianView")

def replace_with_point(expr_str):
    # Регулярное выражение для поиска нужных скобок
    pattern = r'''
        (?<!\w)           # Отрицательный просмотр назад: перед скобкой не должно быть символа слова
        \(                # Открывающая круглая скобка
        \s*               # Возможные пробелы после открывающей скобки
        [^(),]+           # Первый элемент (не содержит скобок или запятых)
        \s*               # Возможные пробелы
        ,                 # Запятая, разделяющая элементы
        \s*               # Возможные пробелы
        [^(),]+           # Второй элемент (не содержит скобок или запятых)
        \s*               # Возможные пробелы перед закрывающей скобкой
        \)                # Закрывающая круглая скобка
    '''
    
    # Замена: вставляем 'Point' перед найденными скобками
    return re.sub(pattern, r'Point\g<0>', expr_str, flags=re.VERBOSE)

 # Функция для извлечения идентификаторов из выражения
def extract_identifiers(expression):
    # Ищем идентификаторы, которые могут содержать фигурные скобки как часть имени
    # Это включает имена типа "M_{1}", "u_{1}" и т.д.
    identifier_pattern = r'[a-zA-Z_][a-zA-Z0-9_]*\{[^}]*\}|[a-zA-Z_][a-zA-Z0-9_]*'
    
    # Находим все совпадения
    identifiers = set()
    for match in re.finditer(identifier_pattern, expression):
        ident = match.group(0)
        # Проверяем, что это не часть числа и не оператор
        if not re.match(r'^\d', ident) and ident not in ['and', 'or', 'not']:
            identifiers.add(ident)
    
    # Также ищем идентификаторы с специальными символами, которые могут быть пропущены
    special_pattern = r'[a-zA-Z_][a-zA-Z0-9_{}\'&°]*'
    for match in re.finditer(special_pattern, expression):
        ident = match.group(0)
        # Исключаем чисто числовые значения и операторы
        if not re.match(r'^\d', ident) and ident not in ['and', 'or', 'not']:
            identifiers.add(ident)
    
    # Исключаем ключевые слова Python и стандартные функции
    python_keywords = {'and', 'as', 'assert', 'break', 'class', 'continue', 
                      'def', 'del', 'elif', 'else', 'except', 'False', 'finally', 
                      'for', 'from', 'global', 'if', 'import', 'in', 'is', 
                      'lambda', 'None', 'nonlocal', 'not', 'or', 'pass', 'raise', 
                      'return', 'True', 'try', 'while', 'with', 'yield'}
    
    math_functions = {'sin', 'cos', 'tan', 'sqrt', 'log', 'ln', 'exp', 'abs', 
                     'round', 'floor', 'ceil', 'min', 'max', 'sum'}
    
    excluded = python_keywords.union(math_functions)
    return identifiers - excluded

def convert_ggb_expr_to_python(constr, expr_str, expr_type=None):    
    # Извлекаем все идентификаторы из выражения
    identifiers = extract_identifiers(expr_str)
    # print("identifiers:", identifiers)
    
    # Добавляем новые идентификаторы в name_mapping, если они еще не там
    for identifier in identifiers:
        if identifier not in constr.name_mapping:
            constr.get_normalized_name(identifier)
    
    # Заменяем все имена согласно name_mapping
    for original, normalized in constr.name_mapping.items():
        expr_str = re.sub(r'\b' + re.escape(original), normalized, expr_str)
        
    # Проверяем, является ли выражение простым вектором в формате (x, y)
    vector_match = re.match(r'^\s*\(\s*([-\d.]+)\s*,\s*([-\d.]+)\s*\)\s*$', expr_str)
    if vector_match and expr_type == "vector":
        x, y = vector_match.groups()
        return f"Vector(Point({x}, {y}))"
    
    expr_str = replace_with_point(expr_str)
    
    # Заменяем квадратные скобки на круглые
    expr_str = expr_str.replace('[', '(').replace(']', ')')
    
    # Заменяем функции CamelCase на snake_case
    pattern = r'\b([A-Za-z_][A-Za-z0-9_]*)\('
    def camel_to_snake(match):
        name = match.group(1)
        s = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
        return s + '('
    
    expr_str = re.sub(pattern, camel_to_snake, expr_str)
    
    # Заменяем градусы
    pattern_deg = r'(\d+)°'
    expr_str = re.sub(pattern_deg, r'AngleSize("\1°")', expr_str)
    
    return expr_str

def is_number(s):
    try:
        float(s)
        return True
    except ValueError:
        return False

def is_angle_degrees(s):
    try:
        assert(s[-1] == '°')
        float(s[:-1])
        return True
    except:
        return False

def is_simple_value(s):
    if s is None:
        return True
    if is_number(s):
        return True
    if is_angle_degrees(s):
        return True
    if re.match(r'^[a-zA-Z_][a-zA-Z0-9_]*$', s):
        return True
    return False

def get_alpha_from_point_and_circle(point, circle):
    """Вычисляет alpha для точки на окружности"""
    vector = point.a - circle.c
    theta = np.arctan2(vector[1], vector[0])
    return theta % (2*np.pi)

def get_alpha_from_point_and_line(point, line):
    """Вычисляет alpha для точки на прямой"""
    base_point = line.c * line.n
    t = np.dot(point.a - base_point, line.v)
    return t

def get_alpha_from_point_and_segment(point, segment):
    """Вычисляет alpha для точки на отрезке"""
    A = segment.end_points[0]
    B = segment.end_points[1]
    AB = B - A
    AP = point.a - A
    alpha = np.dot(AP, AB) / np.dot(AB, AB)
    return alpha

def parse_constr(constr: Construction, constr_xelem: XElement, debug = False):
    xelems_left_to_pass = 0
    fixed_element = False
    alpha_input = None
    name_mapping = {}
    style = {}

    for xelem in constr_xelem:                    
        if xelem.tag == "element":                    
            name = xelem.attrib['label']
            type = xelem.attrib['type']

            name_mapping[name] = constr.get_normalized_name(name)
            name = name_mapping[name]
            
            style[name] = {}

            elem = xelem.find("decoration")
            if elem is not None: 
                style[name]['lines'] = int(elem.attrib['type'])
                if type == 'angle':
                    style[name]['lines'] += 1
                
            elem = xelem.find("show")
            if elem is not None: 
                style[name]['show_element'] = (elem.attrib['object'] == 'true')
                style[name]['show_label'] = (elem.attrib['label'] == 'true')
            
            elem = xelem.find("labelMode")
            if elem is not None:
                caption = xelem.find("caption")
                if (elem.attrib['val'] == '3') & (caption is not None):
                    style[name]['label'] = '$' + caption.attrib['val'] + '$'
            
            elem = xelem.find("angleStyle")
            if elem is not None:
                if type == 'angle':
                    style[name]['range_type'] = int(elem.attrib['val'])
                    
            elem = xelem.find("pointStyle")
            if elem is not None:
                if type == 'point':
                    style[name]['point_type'] = int(elem.attrib['val'])
                    
            # elem = xelem.find("pointSize")
            # if elem is not None:
            #     if type == 'point':
            #         style[name]['size'] = float(elem.attrib['val']) * 0.01 * 2.8346456692916

            elem = xelem.find("labelOffset")
            if elem is not None: 
                x, y = float(elem.attrib['x']), float(elem.attrib['y'])
                if type == "angle":
                    style[name]['offset'] = [x/100, -y/100]
                else:
                    style[name]['offset'] = [0.1 + x/100, 0.11 - y/100]
            else:
                if type == "angle":
                    style[name]['offset'] = [0, 0]
                else:
                    style[name]['offset'] = [0.1, 0.11]

            elem = xelem.find("arcSize")
            if elem is not None: 
                if xelem.attrib['type'] == 'angle':
                    style[name]['r_offset'] = (float(elem.attrib['val']) - 30) / 150
                
            elem = xelem.find("objColor")
            if elem is not None:
                r, g, b, a = int(elem.attrib['r']), int(elem.attrib['g']), int(elem.attrib['b']), float(elem.attrib['alpha'])
                if xelem.attrib['type'] in ['angle', 'polygon', 'arc', 'conic']: #, 'point']:
                    style[name]['fill'] = rgb_to_hex(r, g, b)
                    style[name]['fill_opacity'] = a
                lineStyle = xelem.find("lineStyle")
                if lineStyle is not None:
                    thick, tt = lineStyle.attrib['thickness'], lineStyle.attrib['type']
                    op = float(lineStyle.attrib['opacity']) if 'opacity' in lineStyle.attrib else 255
                    if xelem.attrib['type'] in ['angle', 'segment', 'arc', 'conic', 'vector']:
                        style[name]['stroke'] = rgb_to_hex(r, g, b)
                        #style[name]['stroke_opacity'] = op / 255
                        style[name]['stroke_width'] = thick
                        if int(tt) > 0: style[name]['stroke_dash'] = 0.65
                        
        if xelems_left_to_pass:
            xelems_left_to_pass -= 1
            continue

        if xelem.tag == "expression":
            name = xelem.attrib['label']
            expr = xelem.attrib['exp']
            expr_type = xelem.attrib.get('type', None)
            name_mapping[name] = constr.get_normalized_name(name)
            converted_expr = convert_ggb_expr_to_python(constr, expr, expr_type)

            if debug:
                print(f"Expression found: {name} = {converted_expr}")
                
            if expr.find('=') < 0:
                putCode(constr, f"{name_mapping[name]} = {converted_expr}", debug=debug)
            else:
                print(f"{expr_type}: {expr} >> {name}")
                print(f"WARNING: parser has no implementation for expression '{expr}'\n")
            
            xelems_left_to_pass = 1
            continue
        
        if xelem.tag == "command":            
            comm_name = xelem.attrib["name"]                        
            input_xelem, output_xelem = xelem.find("input"), xelem.find("output")
            inputs, outputs = list(input_xelem.attrib.values()), list(output_xelem.attrib.values())
            outputs = [constr.get_normalized_name(output) for output in outputs]
            for i, inp in enumerate(inputs):
                if is_number(inp) or is_angle_degrees(inp):
                    inputs[i] = inp 
                else:
                    inputs[i] = constr.get_normalized_name(inp)
            
            # Обработка выражений во входных параметрах
            new_inputs = []
            for inp in inputs:
                if is_simple_value(inp):
                    new_inputs.append(inp)
                else:
                    # Создаем временную переменную для выражения
                    temp_name = constr.add_new_phantom()
                    converted_expr = convert_ggb_expr_to_python(constr, inp)
                    putCode(constr, f"{temp_name} = {converted_expr}", debug=debug)
                    new_inputs.append(temp_name)
            
            command = Command(comm_name, new_inputs, outputs)
            constr.add(command)
            constr.apply(command, debug = debug)
            
            if comm_name == "Point":
                if len(inputs) == 1:
                    if debug: print(f'Point FIXED by {inputs[0]}')
                    fixed_element = True
            
                    input0 = constr.element(inputs[0]).data if constr.element(inputs[0]) else None
                    if isinstance(input0, (Circle, Line, Segment)):
                        alpha_input = input0
            
            xelems_left_to_pass = len(output_xelem.attrib) if not alpha_input else 0
                        
            continue
        
        # Here xelem has to be a commandless point or numeric (Var)
        
        if xelem.tag == "element":
            name = name_mapping[xelem.attrib["label"]]
            if xelem.attrib["type"] == "point":
                coords = list(xelem.find("coords").attrib.values())
                coords.pop(-1) #  removing z coordinate
                alpha = None
                if alpha_input:
                    if isinstance(alpha_input, Circle):
                        alpha = get_alpha_from_point_and_circle(Point(coords), alpha_input)
                    elif isinstance(alpha_input, Segment):
                        alpha = get_alpha_from_point_and_segment(Point(coords), alpha_input)
                    elif isinstance(alpha_input, Line):
                        alpha = get_alpha_from_point_and_line(Point(coords), alpha_input)
                
                if alpha is not None: 
                    fixed_element = False
                    constr.element(name).data = Point([float(x) for x in coords])
                    constr.element(name).alpha = alpha
                else:
                    constr.add(Element(name, Point([float(x) for x in coords]), fixed = fixed_element))
                    
                fixed_element = False
                alpha_input = None
                continue
            if xelem.attrib["type"] == "numeric":
                value_xelem = xelem.find("value")
                constr.add(Var(name, float(value_xelem.attrib["val"])))
                continue
            if xelem.attrib["type"] == "angle":
                value_xelem = xelem.find("value")
                constr.add(Var(name, AngleSize(float(value_xelem.attrib["val"]))))
                continue
        
        #raise ElementTree.ParseError(f"Unexpected XElement met:\n\t<{xelem.tag}>, {xelem.attrib}")
        print(f"Unexpected XElement met:\n\t<{xelem.tag}>, {xelem.attrib}")

    constr.rebuild(debug = debug)

    #styling elements
    for name in style:
        if constr.element(name) is None: continue 
        if isinstance(constr.element(name), Var): continue 
        for key in style[name]:
            if debug: print(f'STYLE >> {name} >> {key} = {style[name][key]}')
            if key == 'show_element':
                if not constr.element(name):
                    print(f"element {name} is None")
                constr.element(name).visible = style[name][key]
                continue
            constr.element(name).style[key] = style[name][key]

def FloatOrNone(txt):
    return float(txt) if txt is not None else None

def parse_view(constr: Construction, view_xelem: XElement, debug = False):
    constr.style['view'] = {}
    
    for xelem in view_xelem:            
        if xelem.tag == "size":
            constr.style['view']['width'], constr.style['view']['height'] = FloatOrNone(xelem.attrib['width']), FloatOrNone(xelem.attrib['height'])

        if xelem.tag == "coordSystem":
            constr.style['view']['xZero'], constr.style['view']['yZero'] = FloatOrNone(xelem.attrib['xZero']), FloatOrNone(xelem.attrib['yZero'])
            constr.style['view']['scale'] = FloatOrNone(xelem.attrib['scale'])       

def load(constr: Construction, ggb_path: str, debug = False): # -> Construction:
    if debug: print("--------------------------------\n[[loading from ggb]]:\n")
    constr_xelem, view_xelem = get_xelems(ggb_path)
    # if debug: print(constr_xelem, view_xelem)
    parse_constr(constr, constr_xelem, debug = debug)
    parse_view(constr, view_xelem, debug = debug)



from setuptools import setup, find_packages

def readme():
    with open('README.md', 'r') as f:
        return f.read()

setup(
    name='animageo',
    version='0.1.11',
    author='ivaleo',
    author_email='ivaleotion@gmail.com',
    description='Tools for using GeoGebra construction, processing with Manim animation and exporting to SVG and MP4',
    long_description=readme(),
    long_description_content_type='text/markdown',
    url='http://animageo.ru/',
    packages=find_packages(),
    entry_points={
        'console_scripts': ['animageo=animageo.__main__:main'],
    },
    install_requires=[
        'requests>=2.25.1',
        'numpy>=1.26.0',
        'manim>=0.19.0',
        'pycairo>=1.0.0',
        'lark>=1.1.5'
    ],
    classifiers=[
        'Programming Language :: Python :: 3.13',
        'Operating System :: OS Independent'
    ],
    license='MIT',
    keywords='geometry dynamic geogebra manim animation drawing svg mp4 python',
    python_requires='>=3.9'
)


from .lib_vars import *
from .lib_elements import *
from .lib_commands import *
import itertools
from collections import deque

#--------------------------------------------------------------------------

def appendIfNew(elem, elemArray):
    for el in elemArray:
        if elem == el: return
    
    elemArray.append(elem)
    
def normalize_name(name):
    # Заменяем HTML-сущности
    name = name.replace("&apos;", "_prime")
    # Заменяем нижние индексы в фигурных скобках
    name = re.sub(r'{(\w+)}', r'\1', name)
    # Заменяем любые другие недопустимые символы на подчеркивание
    name = re.sub(r'[\']', "_Prime", name)
    # Убедимся, что имя не начинается с цифры
    if name[0].isdigit():
        name = 'var_' + name
    return name

class Construction:
    def __init__(self):
        self.phantoms = {}   #example: {'_1': Var, '_2': Point, '_3': [Point, float, '_2']}
        self.vars = []
        self.elements = [
            Element("xAxis", Line((0, 1), 0), visible=False),
            Element("yAxis", Line((1, 0), 0), visible=False)
        ]
        self.commands = []
        self.state = {} #статус элементов и связи (state[elem_name] = {level, inputs, outputs, input_commands, built})
        self.name_mapping = {} #перевод сложных имен в нормализованные: t_{1} -> t_1
        
        self.style = {}

    def __repr__(self):
        str_out = "--------------------------------\n[[construction]]:\n"

        str_out += "\n[{} phantoms]:".format(len(self.phantoms)) + '\n'
        for key in self.phantoms: str_out += str(key) + ': ' + str(self.phantoms[key]) + '\n'

        str_out += "\n[{} vars]:".format(len(self.vars)) + '\n'
        for var in self.vars: str_out += str(var) + '\n'
       
        str_out += "\n[{} elements]:".format(len(self.elements)) + '\n'
        for element in self.elements: str_out += str(element) + '\n'
        
        str_out += "\n[{} commands]:".format(len(self.commands)) + '\n'
        for command in self.commands: str_out += str(command) + '\n'

        str_out += "\n................................\n"
        str_out += "[[state]]:\n"

        state = []
        for name in self.state:
            level = self.state[name]['level']
            while level >= len(state): state.append([])
            state[level].append(name)

        for level in range(len(state)):
            str_out += f"\n[level {level}]:\n"
            for name in state[level]:
                str_out += name + ('' if self.state[name]['built'] else '*') + "\t"

        str_out += "\n--------------------------------\n"
        return str_out

    def new_repr(self):
        return "--------------------------------\n[[construction]]:\n" + \
            f"\n[{len(self.vars)} vars]:\n" + '\n'.join(map(str, self.vars)) + "\n" + \
                f"\n[{len(self.elements)} elements]:\n" + '\n'.join(map(str, self.elements)) + "\n" + \
                    f"\n[{len(self.commands)} commands]:\n" + '\n'.join(map(str, self.commands)) + "\n" + \
                        "--------------------------------\n"

    def add_new_phantom(self):
        num = 1
        while ('_' + str(num)) in self.phantoms: num += 1
        key = '_' + str(num)
        self.phantoms[key] = None
        return key
    
    def get_normalized_name(self, name):
        normalized_name = normalize_name(name)
        self.name_mapping[name] = normalized_name
        return normalized_name

    def add(self, obj):
        if isinstance(obj, Var): 
            self.vars.append(obj)
            if obj.name not in self.state:
                self.state[obj.name] = { 'level': 0, 'inputs': [], 'outputs': [], 'input_commands': [], 'built': True }
        elif isinstance(obj, Element): 
            self.elements.append(obj)
            if obj.name not in self.state:
                self.state[obj.name] = { 'level': 0, 'inputs': [], 'outputs': [], 'input_commands': [], 'built': True }
        elif isinstance(obj, Command): 
            self.commands.append(obj)                    
            self.updateStateLevels(obj)

    def update(self, name, data, log = None):
        if isinstance(data, str):
            data_prepared = self.dataByStr(data)
            if data_prepared: data = data_prepared
        
        obj = self.objectByName(name)
        if obj is None:
            if name not in self.state:
                self.state[name] = { 'level': 0, 'inputs': [], 'outputs': [], 'input_commands': [], 'built': False }
            if isinstance(data, (Point, Line, Angle, Polygon, Circle, Vector)):
                self.add(Element(name, data, visible = (name[0] != '_')))
                self.state[name]['built'] = True
                if log is not None: log[name] = True
            elif isinstance(data, (int, float, bool, Measure, AngleSize)):
                self.add(Var(name, data))
                self.state[name]['built'] = True
                if log is not None: log[name] = True
            elif data is not None:
                print(f"Construction.update({name}) ERROR: element data type {type(data)} update has no realization")
            else:
                self.add(Element(name, None, visible = (name[0] != '_')))
                self.state[name]['built'] = False
                if log is not None: log[name] = True
        else:
            if isinstance(obj, Var) or isinstance(obj, Element):
                for output in self.state[name]['outputs']:
                    self.state[output]['built'] = False
                # ?здесь нужно ли проверить, что объект не имеет предшедствующих зависимостей
                if not isinstance(obj.data, type(data)):
                    if isinstance(obj, Var):
                        if isinstance(obj.data, AngleSize) and isinstance(data, (int, float)):
                            data = AngleSize(data)
                        elif isinstance(obj.data, (int, float)) and isinstance(data, AngleSize):
                            data = data.value
                    elif data is not None and obj.data is not None:
                        print(f"Construction.update('{name}') ERROR: update data type is uncompatible: {obj.data} != {data}")
                obj.data = data
                self.state[name]['built'] = True
                if log is not None: log[name] = True
            elif isinstance(obj, Command):
                print(f"Construction.update('{name}') ERROR: no implementation for command update")

    def _updateStateLevel(self, name, level):
        queue = deque()
        queue.append((name, level))
        
        while queue:
            current_name, current_level = queue.popleft()
            
            if self.state[current_name]['level'] < current_level:
                self.state[current_name]['level'] = current_level
                
                for output in self.state[current_name]['outputs']:
                    next_level = current_level + 1
                    if next_level > self.state[output]['level']:
                        queue.append((output, next_level))

    def sortCommands(self):
        #! todo поправить state (расположение элементов по уровням!) после сортировки
        
        i_first = -1
        i_last = len(self.commands) - 1
        for _ in range(len(self.commands)):
            cmd_last = self.commands[i_last]
            i = 0
            while (i < i_last) and (i_first < 0):
                cmd_first = self.commands[i]
                for input, output in itertools.product(cmd_first.inputs, cmd_last.outputs):
                    if input == output:
                        i_first = i
                        break
                i += 1
                    
            if i_first >= 0:
                self.commands.pop(i_last)
                self.commands.insert(i_first, cmd_last)
            else:
                i_last -= 1
            i_first = -1
        
        #! todo нужно в конце еще один раз проверить, если снова есть нарушенный порядок, значит, это неизбежно и есть циклы зависимостей между командами

    def updateStateLevels(self, command):
        for output in command.outputs:
            if is_number(output): raise Exception(f'output could not be a number: {output}')
            if is_angle_degrees(output): raise Exception(f'output could not be an angle_size: {output}')    
            if output not in self.state:
                self.state[output] = { 'level': 0, 'inputs': [], 'outputs': [], 'input_commands': [], 'built': False }
            
            self.state[output]['built'] = False
            appendIfNew(command, self.state[output]['input_commands'])
            
            for input in command.inputs:
                if is_number(input): continue
                if is_angle_degrees(input): continue
                if input not in self.state:
                    self.state[input] = { 'level': 0, 'inputs': [], 'outputs': [], 'input_commands': [], 'built': False }

                appendIfNew(output, self.state[input]['outputs'])
                appendIfNew(input, self.state[output]['inputs'])
                
                self._updateStateLevel(output, max(self.state[output]['level'], self.state[input]['level'] + 1))


    def updateCommand(self, nameCommand, inputs, outputs = None):
        command = None if outputs is None else self.commandByElementName(outputs[0])
        if command is not None:
            command.name = nameCommand
            command.inputs = inputs
            command.outputs = outputs
            self.updateStateLevels(command)
        else:
            self.add(Command(nameCommand, inputs, outputs))

    def element(self, name: str): #) -> Element | None:
        result = list(filter(lambda elem: elem.name == name, self.elements))
        return result[0] if result else None

    def var(self, name: str): #) -> Var | None:
        result = list(filter(lambda var: var.name == name, self.vars))
        return result[0] if result else None

    def objectByName(self, name: str): #) -> Element | Var | None:
        for elem in self.elements:
            if elem.name == name: return elem
        for var in self.vars:
            if var.name == name: return var
        for key in self.phantoms:
            if key == name: return self.phantoms[key]   
        return None

    def commandByElementName(self, name: str): #) -> Command | None:
        result = list(filter(lambda comm: name in comm.outputs, self.commands))
        return result[0] if result else None

    def dataByStr(self, text):
        obj = self.objectByName(text)
        if obj is not None: return obj.data

        if is_number(text): return float(text)
        if is_angle_degrees(text): return AngleSizeFromStr(text)
        if is_boolean(text): return boolean(text)

        #raise Exception("Not found object(s) '{}' or not processing".format(text))
        return None

    def prepareInputs(self, command):
        for i in range(len(command.inputs)):
            if isinstance(command.inputs[i], str):
                command.inputs[i] = self.dataByStr(command.inputs[i])

    def rebuild(self, debug = False, full = False):
        if debug: print('\n""""""""""""""""""""""""""""""""')
        
        log = {}
        if full:
            for command in self.commands: self.apply(command, debug, log = log)
        else:
            state = []
            for name in self.state:
                level = self.state[name]['level']
                while level >= len(state): state.append([])
                state[level].append(name)

            for level in range(len(state)):
                for name in state[level]:
                    if not self.state[name]['built']:
                        for command in self.state[name]['input_commands']:
                            self.apply(command, debug = debug, log = log)
        
        return log

    def copy(self, command):
        assert(isinstance(command, Command))
        command_copy = Command(command.name, list(command.inputs).copy(), list(command.outputs).copy())
        return command_copy

    def apply(self, command_original, debug = False, log = None):
        command = self.copy(command_original)            
        self.prepareInputs(command)
        input_data = [obj.data if hasattr(obj,"data") else obj for obj in command.inputs]
        
        # специальный случай для Point: если у Point определено положение с помощью коэфф alpha, то он передается в команду как дополнительный параметр
        if len(command.outputs) == 1 and self.element(command.outputs[0]):
            if isinstance(self.element(command.outputs[0]).data, Point):
                if self.element(command.outputs[0]).alpha is not None:
                    input_data.append(self.element(command.outputs[0]).alpha)
        
        # специальный случай для Intersect(Circle, ...): если уже были точки до этого, то используется порядок точек
        points_order = None
        if command.name == 'Intersect':
            circ, circ_name = None, None
            obj2, obj2_name = None, None
            for i, inp in enumerate(command_original.inputs):
                if isinstance(input_data[i], Circle):
                    circ, circ_name = input_data[i], inp.name if hasattr(inp,"name") else inp
                elif i < 2:
                    obj2, obj2_name = input_data[i], inp.name if hasattr(inp,"name") else inp
            
            if circ:
                points_order = []
                for name in self.state[circ_name]['inputs']:
                    elem = self.element(name)
                    if elem is not None:
                        if isinstance(elem.data, Point): points_order.append(elem.data)
                if obj2 is not None:
                    for name in self.state[obj2_name]['inputs']:
                        elem = self.element(name)
                        if elem is not None:
                            if isinstance(elem.data, Point): points_order.append(elem.data)
                            
                # for circ_output in self.state[circ_name]['outputs']:
                #     if self.element(circ_output):
                #         p = self.element(circ_output).data
                #         if isinstance(p, Point): points_order.append(p)

                if (len(points_order) > 0) and is_number(input_data[-1]):
                    input_data.append(points_order)
        
        f = command.func()

        if f is not None:
            # основной вызов расчетной команды
            output_data = f(*input_data)
            if not isinstance(output_data, list): output_data = [output_data]
            if debug:
                str_inputs = [obj.name if hasattr(obj,"name") else obj for obj in command_original.inputs]
                print(f"{f.__name__}: {str_inputs} >> {command_original.outputs}")
                print(f"    {input_data}")
                print(f"    {output_data}")

            # здесь идет проверка выходных данных output_data и запись соответствующих данных в command.outputs
            for i in range(len(output_data)):
                if (i < len(command.outputs)): #and (output_data[i] is not None):
                    if self.element(command.outputs[i]) is not None:
                        if not self.element(command.outputs[i]).fixed:
                            self.update(command.outputs[i], output_data[i], log = log)
                    else:
                        self.update(command.outputs[i], output_data[i], log = log)
        else:
            str_inputs = [obj.name if hasattr(obj,"name") else obj for obj in command_original.inputs]
                    
            has_none_attr = False
            for inp in input_data: 
                if inp is None: 
                    has_none_attr = True
                    break
            
            if has_none_attr:
                if debug:
                    print(f"{strFullCommand(command.name, command.inputs)}: {str_inputs} >> {command_original.outputs}")
                    print(f"    {input_data}")
                    print(f"WARNING: undefined arguments in '{command.name}'\n")
            else:
                str_types = [type(x).__name__ for x in input_data]
                print(f"{strFullCommand(command.name, command.inputs)}: {str_inputs} >> {command_original.outputs}")
                print(f"    {input_data}")
                print(f"WARNING: no implementation for '{command.name}' and arguments {str_types}\n")
                
            for i in range(len(command.outputs)):
                if self.element(command.outputs[i]) is not None:
                    if not self.element(command.outputs[i]).fixed:
                        self.update(command.outputs[i], None, log = log)
            

def is_number(s):
    try:
        float(s)
        return True
    except:
        return False
    
def is_boolean(s):
    try:
        assert(str(s).lower() in ['true', 'false'])
        return True
    except:
        return False
    
def boolean(s):
    return (str(s).lower() == 'true')

def is_angle_degrees(s):
    try:
        assert(s[-1] == '°')
        float(s[:-1])
        return True
    except:
        return False